from pathlib import Path
import re

import numpy as np
import pandas as pd
import xarray

import pytsg.parse_tsg


def load_tsg(
    directory, spectra="NIR", image=True, subsample_image=10, index_coord="sample"
):
    """
    Load a TSG dataset.

    Parameters
    ----------
    directory : str | pathlib.Path
        Directory of the TSG datset to load.
    spectra : str
        Which spectra to load by default, NIR or TIR.
    image : bool
        Whether to load the high-resolution RGB imagery.
    subsample_image : int
        Subsampling factor for the high-resolution RGB imagery.
        Default of 10 returns a 100x reduction in image size/
        1% of all pixels.
    index_coord : str
        Index coordinate to use for the dataset.
        Using "depth" requires some post-processing and dropping duplicates.

    Returns
    -------
    xarray.Dataset
        Dataset containing the spectra and assocaited data.
    """
    directory = Path(directory)
    tsgdata = pytsg.parse_tsg.read_package(directory, read_cras_file=image)
    dataset = tsg_to_xarray(tsgdata, spectra, index_coord=index_coord)
    if image:
        dataset["Image"] = cras_to_dataarray(tsgdata, subsample=subsample_image)
    dataset = reorder_variables(dataset)
    return dataset


def tsg_to_xarray(tsgdata, spectra, index_coord="sample"):
    """
    Load a TSG spectral subset into Xarray.

    Parameters
    ----------
    tsgdata  : pytsg.parse_tsg.TSG
        TSG dataset loaded with pytsg.
    spectra : str
        Which spectra to load by default, NIR or TIR.
    index_coord : str
        Index coordinate to use for the dataset.
        Using "depth" requires some post-processing and dropping duplicates.

    Returns
    -------
    xarray.Dataset
        Dataset containing spectra and band headers.

    Todo
    -----
    * Consider indexing by depth instead of sample, after the fact.
    * Consider dropping Tray, Section, Depth (m) as they're duplicated as indexes.
    * Consider dropping SecDist (mm), TraySamp, SecSamp and NumFeats - they can be calculated.
    """
    spectraldata = getattr(tsgdata, spectra.lower())
    assert hasattr(spectraldata, 'spectra'), "TSG Dataset does not have {} data.".format(
        spectra
    )
    scalar_data = spectraldata.scalars.copy()
    floatvals = scalar_data.select_dtypes(float).columns
    scalar_data[floatvals] = np.where(
        np.isclose(scalar_data.loc[:, floatvals].values, np.finfo("float32").min),
        np.nan,
        scalar_data.loc[:, floatvals.values],
    )
    # could drop emtpy columns but is unlikely to be many
    dataset = scalar_data.set_index(
        pd.Series(scalar_data.index.values, name="sample")
    ).to_xarray()
    dataset.attrs.update(
        {
            ch.name: [(i, v) for i, v in ch.classes.items()]
            for id, ch in spectraldata.classes.items()
        }
    )

    for grp in ["Centre", "Depth", "Width"]:
        arr = (
            dataset[[v for v in dataset.data_vars if re.match(grp + "\d+", v)]]
            .to_array()
            .rename({"variable": "feature"})
        )
        dataset = dataset[
            [v for v in dataset.data_vars if v not in arr.coords["feature"]]
        ]
        arr["feature"] = [f.replace(grp, "") for f in arr["feature"].values]
        arr = xarray.where(arr == 0, np.nan, arr)
        arr.attrs = {}

        dataset[grp + "s"] = arr

    # convert traynames, otherwise occasionally converted to integers
    dataset["Tray"] = dataset["Tray"].astype("<U16")
    # add the spectra, and move it to the top of the variable list
    coords = coords_from_sampleheaders(spectraldata)
    specarr = xarray.DataArray(
        spectraldata.spectra, coords=coords, dims=("sample", "wavelength")
    )
    if index_coord != "depth":
        if tsgdata.lidar is not None:
            profilometer = xarray.DataArray(
                tsgdata.lidar, coords={"sample": specarr.sample.values}
            )
        else:
            profilometer = None
        # alternate method for being able to index on depth for spectral without
        # dropping rows
        # specarr = specarr.set_xindex('holedepth')
    else:
        # remove samples where the depth is a duplicate, and sort by depth
        # to allow depth as an index
        fltr = pd.Series(specarr.holedepth).duplicated().values
        specarr = specarr.sel(sample=~fltr)

        sortidx = np.argsort(specarr.holedepth.values)
        specarr = specarr[sortidx].swap_dims({"sample": "holedepth"})

        if tsgdata.lidar is not None:
            profilometer = xarray.DataArray(
                tsgdata.lidar[~fltr][sortidx],
                coords={"holedepth": specarr.holedepth.values},
            )
        else:
            profilometer = None

    dataset["Spectra"] = specarr
    if profilometer is not None:
        dataset["Lidar"] = profilometer
    dataset = dataset[["Spectra"] + [v for v in dataset.data_vars if v != "Spectra"]]
    return dataset


def coords_from_sampleheaders(spectraldata):
    """
    Turn the sample headers of a TSG spectral subset into coordinates.

    Parameters
    ----------
    spectraldata  : pytsg.parse_tsg.Spectra
        Spectral subset loaded with pytsg.

    Returns
    -------
    coords : dict
        Mapping of coordinate names to values, and in the case of non-index coordinates
        the corresponding index coordinate.
    """
    sampleheaders = spectraldata.sampleheaders.apply(
        pd.to_numeric, errors="ignore"
    ).rename(
        columns={
            "sample": "sample",
            "T": "tray",
            "L": "section",
            "P": "section-part",
            "D": "holedepth",  # holedepth here as we can't easily deal with two depth indexes
            "X": "section-position",
            "H": "hole",
        }
    )
    # note that depths can be duplicated, so would need to be
    # post-processed to be used as an index
    coords = {
        "sample": sampleheaders["sample"].values,
        "wavelength": spectraldata.wavelength,
    }
    coords.update(
        {
            c: ("sample", d.values)
            for c, d in sampleheaders.items()
            if c not in ["sample"]
        }
    )
    return coords


def cras_to_dataarray(tsgdata, subsample=10):
    """
    Get the high resolution imagery from a TSG file, and optionally subsample it
    to a lower resolution.

    Parameters
    ----------
    tsgdata : pytsg.parse_tsg.TSG
        TSG dataset loaded with pytsg.

    Returns
    -------
    xarray.DataArray
        Array containing the RGB imagery.
    """
    depths = np.hstack(
        [
            np.linspace(mn, mx, t.nlines)
            for (t, (mn, mx)) in zip(
                tsgdata.cras.section,
                tsgdata.nir.sampleheaders[["T", "L", "D"]]
                .apply(pd.to_numeric)
                .groupby(["T", "L"])
                .agg(["min", "max"])
                .values,
            )
        ]
    )
    dx = dy = np.median(np.diff(depths[:200]))
    horizontal = np.arange(0, tsgdata.cras.image.shape[1]) * dy
    horizontal -= horizontal.mean()
    cras = xarray.DataArray(
        tsgdata.cras.image,
        coords={"depth": depths, "horizontal": horizontal, "channel": list("RGB")},
    )
    return cras[::subsample, ::subsample]


def reorder_variables(
    ds,
    drop=[],  # ["Tray", "Section", "Depth (m)", "SecDist (mm)", "TraySamp", "SecSamp"],
    patterns=[
        "Grp\d*",
        "Min\d*",
        "Wt\d*",
        "Error\d*",
        "SNR",
        "NIL_Stat",
        "Cust",
        "Bound_Water",
        "Unbound_Water",
    ],
):
    """
    Reorder the variables within an Xarray dataset containing TSG data such that
    it's more easily visually navigated (note this does not persist upon
    serialization).

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to reorder.
    drop : list
        Variables to exclude (typically either duplciated in indexes or easily calculated).
    patterns : list
        List of regex patterns to match column groups of band headers within TSG files.

    Returns
    -------
    ds : xarray.Dataset
        Reordered dataset.
    """
    arrangement = [
        v
        for v in [
            "HoleID",
            "Date",
            "Depth (m)",
            "Tray",
            "Section",
            "Spectra",
            "Image",
            "Lidar",
            "Centres",
            "Depths",
            "Widths",
        ]
        if v in ds.data_vars
    ]

    arrangement += [
        v
        for ptn in patterns
        for v in sorted([v for v in ds.data_vars if re.match(ptn, v)])
    ]

    others = sorted(
        list(
            set(
                [v for v in ds.data_vars if (v not in arrangement)]
                + ["Flags"]
                + [v for v in ds.data_vars if v.lower() == v]
            )
        )
    )
    ds = ds[
        [v for v in arrangement + others if (v not in drop) and (v in ds.data_vars)]
    ]
    return ds



from pathlib import Path

import typer
import xarray

try:
    from tqdm.auto import tqdm
except ImportError:

    def tqdm(iterable):
        return iterable


from .util import Handle
from . import __version__, find_TSG_datasets, load_tsg


logger = Handle(__name__, level="INFO")


app = typer.Typer()


@app.command()
def TSG2zarr(
    tsgdir: Path = typer.Argument(..., help="TSG directory or filepath."),
    output_dir: Path = typer.Option(
        None, help="Output directory for a Zarr data store."
    ),
    spectra: str = typer.Option(
        "NIR",
        help="Whether to load NIR or TIR data. "
        "Note that if a specific .tsg file is specified, this option will effecitvely be ignored.",
    ),
    index_coord: str = typer.Option(
        "Sample", help="Whether to use sample or depth as an index."
    ),
    image: bool = typer.Option(
        False, help="Whether to load the attached image, where available."
    ),
    subsample_image: int = typer.Option(
        10, help="Subsampling factor for imagery data."
    ),
    log: str = typer.Option(
        None,
        help="Whether to use logging, and if so what level (DEBUG, INFO)",
    ),
    zipfile: bool = typer.Option("--zip"
        True,
        help="Whether to zip the Zarr archive upon creation.",
    ),
):
    """
    Convert TSG file(s) to Zarr.
    """
    if log is not None:
        Handle("tsgxr", level=log.upper())
    if (
        not tsgdir.is_dir() and tsgdir.suffix.lower() == ".tsg"
    ):  # pointing to a specific .tsg file
        logger.info("Loading TSG file: {}".format(tsgdir.name))
        spectra = "TIR" if tsgdir.stem.lower().endswith("tir") else "NIR"
        assert tsgdir.exists(), "Specified TSG file does not exist."
        datasets = {tsgdir.stem: tsgdir.parent}
    else:
        logger.info("Loading TSG files from directory: {}".format(tsgdir.name))
        assert tsgdir.exists(), "Specified directory does not exist."
        datasets = find_TSG_datasets(tsgdir)

    if datasets:
        logger.info("Found datasets: {}".format(", ".join(k for k in datasets.keys())))
    else:
        logger.warning("Found no datasets in {}.".format(str(tsgdir.resolve())))

    if datasets:
        for k, d in tqdm(datasets.items()):
            ds = load_tsg(
                d,  
                spectra=spectra.upper(),
                image=image,
                subsample_image=subsample_image,
                index_coord=index_coord,
            )
            name = "_".join([str(ds.coords["hole"][0].values), spectra]) + (
                ".zip" if zipfile else ".zarr"
            )
            # put it in the TSG directory if an output folder is not given
            outdir = output_dir if output_dir is not None else d
            logger.info("Creating Zarr archive {} in {}.".format(name, str(outdir)))
            ds.to_zarr(outdir / name, mode="w")  # overwrite if needed


def version_callback(value: bool):
    if value:
        print(f"tsg-xr Version: {__version__}")
        raise typer.Exit()


@app.callback()
def callback(
    version: bool = typer.Option(
        None, "--version", callback=version_callback, is_eager=True
    ),
):
    """
    tsg-xr: A CLI tool for loading and transforming TSG files.
    """
    pass


if __name__ == "__main__":
    app()




import logging
from pathlib import Path


def rm_tree(pth: Path):
    """Recursively remove a folder."""
    if pth.exists():
        if pth.is_file():
            pth.unlink()
        else:
            for child in pth.iterdir():
                if child.is_file():
                    child.unlink()
                else:
                    rm_tree(child)
            pth.rmdir()


def Handle(
    logger,
    handler_class=logging.StreamHandler,
    formatter="%(asctime)s %(name)s - %(levelname)s: %(message)s",
    level=None,
):
    """
    Handle a logger with a standardised formatting.

    Parameters
    -----------
    logger : :class:`logging.Logger` | :class:`str`
        Logger or module name to source a logger from.
    handler_class : :class:`logging.Handler`
        Handler class for the logging messages.
    formatter : :class:`str` | :class:`logging.Formatter`
        Formatter for the logging handler. Strings will be passed to
        the :class:`logging.Formatter` constructor.
    level : :class:`str`
        Logging level for the handler.

    Returns
    ----------
    :class:`logging.Logger`
        Configured logger.

    Notes
    -----
    Duplicated version of pyrolite.util.log.Handle.
    """
    if isinstance(logger, str):
        logger = logging.getLogger(logger)
    elif isinstance(logger, logging.Logger):
        pass
    else:
        raise NotImplementedError
    logger.propagate = False
    if isinstance(formatter, str):
        formatter = logging.Formatter(formatter)

    active_handlers = [
        i
        for i in logger.handlers
        if isinstance(i, (handler_class))  # not a null handler
    ]
    if active_handlers:
        handler = active_handlers[0]  # use the existing stream handler
    else:
        handler = handler_class()
    handler.setFormatter(formatter)
    if handler not in active_handlers:
        logger.addHandler(handler)
    if level is not None:
        logger.setLevel(getattr(logging, level))
    return logger



import versioneer
from setuptools import setup


with open("README.md", "r") as src:
    LONG_DESCRIPTION = src.read()


setup(
    name="tsgxr",
    description="A tool for reading TSG spectra data into Xarray.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    url="https://github.com/CSIRO-GeoscienceAnalytics/tsg-xr",
    project_urls={
        "Code": "https://github.com/CSIRO-GeoscienceAnalytics/tsg-xr",
        "Issue tracker": "https://github.com/CSIRO-GeoscienceAnalytics/tsg-xr/issues",
    },
    author="Morgan Williams",
    author_email="morgan.williams@csiro.au",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords=["TSG", "spectra"],
    packages=["tsgxr"],
    install_requires=["numpy", "xarray", "zarr", "pandas", "pytsg>=0.2.8", "typer"],
    extras_require={
        "plot": ["matplotlib"],
        "dev": ["pytest", "pytest-cov", "coverage", "versioneer", "black", "twine"],
    },
    tests_require=["pytest", "pytest-cov", "coverage"],
    test_suite="test",
    include_package_data=True,
    license="MIT",
    cmdclass=versioneer.get_cmdclass(),
    entry_points={
        "console_scripts": [
            "tsgxr = tsgxr.cli:app",
        ]
    },
)



from pathlib import Path 

from . import _version

__version__ = _version.get_versions()["version"]

from .read import load_tsg


def find_TSG_datasets(parent_directory):
    """
    Check a directory for subdirectories containing Hylogger TSG datasets.

    Parameters
    ----------
    parent_directory : str | pathlib.Path
        Directory containing Hylogger datasets.

    Returns
    -------
    dict
        Dictionary mapping of hole names to subdirectories containing
        Hylogger datasets (the directory to be passed to `pytsg` or
        `tsg-xr`).
    """
    return {
        fpath.stem.replace("_tsg", ""): fpath.parent
        for fpath in sorted(
            set(list(Path(parent_directory).glob("**/*_tsg.tsg")) + list(Path(parent_directory).glob("**/*_tsg_tir.tsg")))
        )
    }
